@@ -44,8 +44,8 @@ struct _PeasPluginInfo {
   gchar *copyright;
   gchar *website;
   gchar *version;
+  gchar *help_uri;
   guint iage;
-  gboolean visible;
   GHashTable *keys;
 
   gint loaded : 1;
@@ -53,6 +53,8 @@ struct _PeasPluginInfo {
      due to an error loading the plugin module (e.g. for Python plugins
      when the interpreter has not been correctly initializated) */
   gint available : 1;
+
+  guint builtin : 1;
 };
 
 PeasPluginInfo *_peas_plugin_info_new   (const gchar    *filename,
